Require Import SigTS.
Require Import SigComp.
Require Import ReachabilityLogic.
Require Import LogicStructComp.
Require Import Coq.Arith.Compare_dec.
Require Import Omega.
Require Import List.
Import ListNotations.
Require Import SigGcd.
Require Import gcd.
(* main state and transition relation *)
Module  MainMod <: SigTSMod.
  
Inductive Loc := c0 | c1 | c' | c''.

Definition State := (Loc*nat*nat*nat*nat)%type.

Inductive tran : State -> State -> Prop:=
|init : forall x0 y0 x y, 
                tran (c0,x0,y0,x,y) (c1,x0,y0,x0,y0)
|go_up : forall x0 y0 x y, x < y ->
               tran(c1,x0, y0,x,y)(c',x0,y0,x,y)
|loop_up : forall x0 y0 x y, x < y ->
           tran(c',x0, y0,x,y)(c',x0,y0,x,y-x)       
|ret_up :  forall x0 y0 x y, y <= x ->
           tran(c',x0, y0,x,y)(c1,x0,y0,x,y)                      |go_dn : forall x0 y0 x y, y < x ->
               tran(c1,x0, y0,x,y)(c'',x0,y0,x,y)
|loop_dn : forall x0 y0 x y, y < x ->
           tran(c'',x0, y0,x,y)(c'',x0,y0,x-y,y)
|ret_dn : forall x0 y0 x y, x <= y ->
               tran(c'',x0, y0,x,y)(c1,x0,y0,x,y).                                
         
Definition SymState : Type :=  State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

Definition trans := tran.
Definition final := fun s => forall s', ~ trans s s'.
Definition phi := fun s =>
                match (s:State) with (c,_,_,x,y) => c = c1 /\ x = y end.

Lemma final_is_final : forall s, phi s <-> final s.
  Proof.
    intros ((((c,x0),y0),x),y) ; split.
    *intros (Hc & Hf) ;
       intros ((((c',x'0),y'0),x'),y'); intro Htran; subst.
     inversion Htran ; subst ; clear Htran;
       try discriminate ; try omega.
    *intro Hf ;  unfold final,phi in *.
      destruct c; auto.
    + exfalso; apply (Hf ((c1, x0, y0, x0, y0)));  apply init.
    + split ; auto.
      destruct (lt_eq_lt_dec x y) as [[Hxy|Hxy]|Hxy]; auto.
      - exfalso ; apply (Hf (c',x0,y0,x,y)); apply go_up; auto.
      - exfalso ; apply (Hf (c'',x0,y0,x,y)); apply go_dn; auto.
    + exfalso.
      destruct (le_gt_dec y x).
      - apply (Hf (c1, x0, y0, x, y)) ; apply ret_up ; auto.
      - apply (Hf (c',x0,y0,x,y-x)) ; apply loop_up ; auto.
     + exfalso.
        destruct (le_gt_dec x y).
      - apply (Hf (c1, x0, y0, x, y)) ; apply ret_dn ; auto.
      - apply (Hf (c'',x0,y0,x-y,y)) ; apply loop_dn ; auto.      Qed.
  
Definition symTrans(q:SymState) :=
  fun s => exists s', q s' /\ trans s' s.

End MainMod.

(*first component*)
Module FstCompMod <: SigTSMod.
  
Inductive Loc := c1 | c'.

Definition State := (Loc*nat*nat*nat*nat)%type.

Inductive tran : State -> State -> Prop:=
|go : forall x0 y0 x y, x < y ->
               tran(c1,x0, y0,x,y)(c',x0,y0,x,y)
|loop : forall x0 y0 x y, x < y ->
           tran(c',x0, y0,x,y)(c',x0,y0,x,y-x)       
|ret :  forall x0 y0 x y, y <= x ->
           tran(c',x0, y0,x,y)(c1,x0,y0,x,y).                     


Definition SymState : Type :=  State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

Definition trans := tran.
Definition final := fun s => forall s', ~ trans s s'.
Definition phi := fun s =>
     match (s:State) with (c,_,_,x,y) => c = c1 /\ y <= x end.

Lemma final_is_final : forall s, phi s <-> final s.
  Proof.
    intros ((((c,x0),y0),x),y) ; split ; intro Hf.
    *intros ((((c',x'0),y'0),x'),y') Htran.
     destruct Hf ; subst ; inversion Htran ; subst ; clear Htran ;     try discriminate ; try omega. 
   * unfold final,phi in *.
     destruct c.
    + split ; auto.
      destruct (le_gt_dec y x) as [Hxy|Hxy] ; auto.
      exfalso.      
      apply (Hf (c',x0,y0,x,y)); apply go; auto.
    + exfalso.
       destruct (le_gt_dec y x) as [Hxy|Hxy] ; auto.
       - apply (Hf (c1,x0,y0,x,y)); apply ret; auto.
       -  apply (Hf (c',x0,y0,x,y-x)) ; apply loop ; auto. 
Qed.
  
Definition symTrans(q:SymState) :=
  fun s => exists s', q s' /\ trans s' s.

End FstCompMod.

(*second component*)
Module SndCompMod <: SigTSMod.
  
Inductive Loc := c1 | c''.

Definition State := (Loc*nat*nat*nat*nat)%type.

Inductive tran : State -> State -> Prop:=
|go : forall x0 y0 x y, y < x ->
               tran(c1,x0, y0,x,y)(c'',x0,y0,x,y)
|loop : forall x0 y0 x y, y < x ->
           tran(c'',x0, y0,x,y)(c'',x0,y0,x-y,y)       
|ret :  forall x0 y0 x y, x <= y ->
           tran(c'',x0, y0,x,y)(c1,x0,y0,x,y).                     


Definition SymState : Type :=  State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

Definition trans := tran.
Definition final := fun s => forall s', ~ trans s s'.
Definition phi := fun s =>
     match (s:State) with (c,_,_,x,y) => c = c1 /\ x <= y end.

Lemma final_is_final : forall s, phi s <-> final s.
  Proof.
    intros ((((c,x0),y0),x),y) ; split ; intro Hf.
    *intros ((((c',x'0),y'0),x'),y') Htran.
     destruct Hf ; subst ; inversion Htran ; subst ; clear Htran ;     try discriminate ; try omega. 
   * unfold final,phi in *.
     destruct c.
    + split ; auto.
      destruct (le_gt_dec x y) as [Hxy|Hxy] ; auto.
      exfalso.      
      apply (Hf (c'',x0,y0,x,y)); apply go; auto.
    + exfalso.
       destruct (le_gt_dec x y) as [Hxy|Hxy] ; auto.
       - apply (Hf (c1,x0,y0,x,y)); apply ret; auto.
       -  apply (Hf (c'',x0,y0,x-y,y)) ; apply loop ; auto. 
Qed.
  
Definition symTrans(q:SymState) :=
  fun s => exists s', q s' /\ trans s' s.

End SndCompMod.



Module Comp1 <: SigCompMod MainMod  FstCompMod.

Definition loc_map : FstCompMod.Loc -> MainMod.Loc :=
  fun c => match c with
            FstCompMod.c1 => MainMod.c1
           |FstCompMod.c' => MainMod.c'
           end.                      
  
Definition state_map : FstCompMod.State -> MainMod.State :=
fun s => match s with (c,x0,y0,x,y) => (loc_map c,x0,y0,x,y) end.

Lemma state_map_inj : forall s'1 s'2 ,  state_map s'1 = state_map s'2 -> s'1 = s'2.
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Hs.
unfold state_map in Hs.
destruct c,c'; unfold loc_map in * ; injection Hs ; intros ;
  subst; try discriminate ; auto.
Qed.

Lemma subtrans : forall s'1 s'2, FstCompMod.trans s'1 s'2 ->  MainMod.trans (state_map s'1) (state_map s'2).
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Htr.
inversion Htr ; subst ; unfold state_map,loc_map ;
  constructor ; auto.
Qed.


Lemma full_subtrans: forall s1 s2 s'1 s'2, MainMod.trans s1 s2 ->
state_map s'1= s1 ->state_map s'2= s2 ->FstCompMod.trans s'1 s'2.
Proof.
  intros  ((((ca,xa0),ya0), xa),ya)  ((((cb,xb0),yb0), xb),yb) ((((c'a,xa'0),ya'0), xa'),ya')  ((((c'b,xb'0),yb'0), xb'),yb') Htr Hsm.
  destruct c'a, c'b ; unfold state_map, loc_map in *; injection Hsm; intros ; subst ; clear Hsm;  injection H4; intros; subst ; clear H4 ; inversion Htr ; subst; try constructor ; auto.
Qed.


Definition isImage(a: MainMod.State)  := exists a',  a = state_map a'.

Lemma clean_border : forall (s'1:FstCompMod.State) (s2:MainMod.State), 
    MainMod.trans (state_map s'1)  s2 ->
    ~isImage s2 -> FstCompMod.final s'1.
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Htr Him.
inversion Htr ; subst ; clear Htr ; destruct c' ; unfold state_map,loc_map in *; try discriminate.
* exfalso ;apply Him; now exists (FstCompMod.c', x0, y0, x, y).
* exfalso ;apply Him; now exists (FstCompMod.c', x0, y0, x, y'-x).
* exfalso ;apply Him; now exists (FstCompMod.c1, x0, y0, x, y).
* intros ((((c',x'0),y'0), x'),y') Htr ; inversion Htr ; subst; omega.
Qed.


Definition symState_map:FstCompMod.SymState -> MainMod.SymState :=
  fun p s =>
    exists s',  p s' /\ (s = state_map s').

End Comp1.



Module Comp2 <: SigCompMod MainMod  SndCompMod.

Definition loc_map : SndCompMod.Loc -> MainMod.Loc :=
  fun c => match c with
            SndCompMod.c1 => MainMod.c1
           |SndCompMod.c'' => MainMod.c''
           end.                      
  
Definition state_map : SndCompMod.State -> MainMod.State :=
fun s => match s with (c,x0,y0,x,y) => (loc_map c,x0,y0,x,y) end.

Lemma state_map_inj : forall s'1 s'2 ,  state_map s'1 = state_map s'2 -> s'1 = s'2.
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Hs.
unfold state_map in Hs.
destruct c,c'; unfold loc_map in * ; injection Hs ; intros ;
  subst; try discriminate ; auto.
Qed.

Lemma subtrans : forall s'1 s'2, SndCompMod.trans s'1 s'2 ->  MainMod.trans (state_map s'1) (state_map s'2).
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Htr.
inversion Htr ; subst ; unfold state_map,loc_map ;
  constructor ; auto.
Qed.


Lemma full_subtrans: forall s1 s2 s'1 s'2, MainMod.trans s1 s2 ->
state_map s'1= s1 ->state_map s'2= s2 ->SndCompMod.trans s'1 s'2.
Proof.
  intros  ((((ca,xa0),ya0), xa),ya)  ((((cb,xb0),yb0), xb),yb) ((((c'a,xa'0),ya'0), xa'),ya')  ((((c'b,xb'0),yb'0), xb'),yb') Htr Hsm.
  destruct c'a, c'b ; unfold state_map, loc_map in *; injection Hsm; intros ; subst ; clear Hsm;  injection H4; intros; subst ; clear H4 ; inversion Htr ; subst; try constructor ; auto.
Qed.


Definition isImage(a: MainMod.State)  := exists a',  a = state_map a'.

Lemma clean_border : forall (s'1:SndCompMod.State) (s2:MainMod.State), 
    MainMod.trans (state_map s'1)  s2 ->
    ~isImage s2 -> SndCompMod.final s'1.
Proof.
intros ((((c',x'0),y'0), x'),y') ((((c,x0),y0), x),y) Htr Him.
inversion Htr ; subst ; clear Htr ; destruct c' ; unfold state_map,loc_map in *; try discriminate.
* intros ((((c',x'0),y'0), x'),y') Htr ; inversion Htr ; subst; omega.
* exfalso ;apply Him;now exists (SndCompMod.c'', x0, y0, x, y).
* exfalso ;apply Him;now exists (SndCompMod.c'', x0, y0, x'-y, y).
* exfalso ;apply Him; now exists (SndCompMod.c1, x0, y0, x, y).
Qed.


Definition symState_map:SndCompMod.SymState -> MainMod.SymState :=
  fun p s =>
    exists s',  p s' /\ (s = state_map s').

End Comp2.




Module SCM := LogicStructCompMod FstCompMod SndCompMod MainMod Comp1 Comp2.
Module RL := SCM.RL.
Module RL1 := SCM.RL1.
Module RL2 := SCM.RL2.


Definition State := MainMod.State.
Definition State1 := FstCompMod.State.
Definition State2 := SndCompMod.State.


Definition SymState := MainMod.SymState.
Definition SymState1 := FstCompMod.SymState.
Definition SymState2 := SndCompMod.SymState.


Definition symState_map1 := Comp1.symState_map.
Definition symState_map2 := Comp2.symState_map.



Module GcdMod <: SigGcdMod.

Definition gcd := gcd.gcd. 

Lemma gcd_lt : forall x y, x < y -> gcd x (y -x) = gcd x y.
Proof.
  intros x y Hxy.
  generalize (gcd_is_gcd  x y) ; intro Hgxy.
  generalize (gcd_is_gcd  x (y - x)) ; intro Hgxy'.
  rewrite gcd_minus in Hgxy; try omega.
  eapply gcd_unique ; eauto.
Qed.

Lemma gcd_eq : forall x y, x = y -> x = gcd x y.
Proof.
  intros x y Heq ; subst.
  generalize (gcd_is_gcd  y y) ; intro Hgyy.
  generalize (gcd_refl y) ; intro Hgyy'.  
  eapply gcd_unique ; eauto.
Qed.

Lemma gcd_gt : forall x y, y < x -> gcd (x - y) y = gcd x y.
Proof.
  intros x y Hxy.
  generalize (gcd_is_gcd  x y) ; intro Hgxy.
  generalize (gcd_is_gcd  (x-y) y) ; intro Hgxy'.
  apply gcd_sym in Hgxy'.
  
  eapply gcd_unique ; eauto.
  rewrite <- gcd_minus ; try omega.
   apply gcd_sym ; assumption.
Qed.

Lemma gcd_sym  : forall x y,  gcd x y = gcd y x.
Proof.
  intros x y.
   generalize (gcd_is_gcd  x y) ; intro Hgxy. 
   generalize (gcd_is_gcd  y x) ; intro Hgyx.
   apply gcd_sym in Hgyx.
   eapply gcd_unique ; eauto.
Qed.

End GcdMod.  

Module ProofMod (Gcd : SigGcdMod).
  Import Gcd.



Definition lhs_main : SymState :=
  fun (s : State) =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c0 /\ 0 < x0 /\ 0 < y0
    end.



Definition rhs_main : SymState :=
  fun (s : State) =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c1 /\ x = y /\ x = gcd x0 y0
    end.




Lemma main_aux1 : MainMod.imp (MainMod.cnj lhs_main MainMod.final) MainMod.bot.
Proof.
unfold MainMod.imp, MainMod.cnj, lhs_main, MainMod.bot.
  intros  ((((c,x0),y0), x),y) ((Hc & Hx0 & Hy0) & Hfin) ; subst. 
  rewrite <- MainMod.final_is_final in Hfin; destruct Hfin;
  discriminate.
Qed.

Definition main_lt :  SymState :=
  fun (s : State) =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ x < y
    end.


Definition main_gt :  SymState :=
  fun (s : State) =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ y < x
    end.



Definition rhs_main_1 : SymState1 :=
  fun (s : State1) =>
    match s with (c,x0,y0,x,y) =>
         c = FstCompMod.c1 /\ x = y /\ x = gcd x0 y0
    end.

Definition main_lt_1 :  SymState1 :=
  fun (s : State1) =>
    match s with (c,x0,y0,x,y) =>
         c = FstCompMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ x < y
    end.


Definition main_gt_1 :  SymState1 :=
  fun (s : State1) =>
    match s with (c,x0,y0,x,y) =>
         c = FstCompMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ y < x
    end.



Definition rhs_main_2 : SymState2 :=
  fun (s : State2) =>
    match s with (c,x0,y0,x,y) =>
         c = SndCompMod.c1 /\ x = y /\ x = gcd x0 y0
    end.

Definition main_lt_2 :  SymState2 :=
  fun (s : State2) =>
    match s with (c,x0,y0,x,y) =>
         c = SndCompMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ x < y
    end.


Definition main_gt_2 :  SymState2 :=
  fun (s : State2) =>
    match s with (c,x0,y0,x,y) =>
         c = SndCompMod.c1 /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ y < x
    end.

Lemma main_lt_rhs_aux1 :
  SCM.weaker (false, RL.rlf main_lt rhs_main)
    (SCM.bRlf_map1 (false, RL1.rlf main_lt_1 rhs_main_1)).
Proof.
  unfold SCM.weaker,   main_lt, rhs_main, main_lt_1, rhs_main_1,
  SCM.bRlf_map1, SCM.Rlf_map1, SCM.symState_map1,
  Comp1.symState_map, Comp1.state_map ; cbn; repeat split ;
    unfold MainMod.imp.
* intros  ((((c,x0),y0), x),y) (Hc & Hgcd & Hx0 & Hy0 & Hxy) ;
   subst.
   exists (FstCompMod.c1,x0,y0,x,y); repeat split; auto.
* intros  ((((c,x0),y0), x),y)
          ( ((((c',x'0),y'0), x'),y') &(Hc' & Hxy & Hgcd) & Heq);
    injection Heq ; intros ; subst; repeat split.
Qed.

Lemma main_gt_rhs_aux1 :
  SCM.weaker (false, RL.rlf main_gt rhs_main)
          (SCM.bRlf_map2 (false, RL2.rlf main_gt_2 rhs_main_2)).
Proof.
  unfold SCM.weaker,   main_gt, rhs_main, main_gt_2, rhs_main_2,
  SCM.bRlf_map2, SCM.Rlf_map2, SCM.symState_map2,
  Comp2.symState_map, Comp2.state_map ; cbn; repeat split ;
    unfold MainMod.imp.
* intros  ((((c,x0),y0), x),y) (Hc & Hgcd & Hx0 & Hy0 & Hxy) ;
   subst.
   exists (SndCompMod.c1,x0,y0,x,y); repeat split; auto.
* intros  ((((c,x0),y0), x),y)
          ( ((((c',x'0),y'0), x'),y') &(Hc' & Hxy & Hgcd) & Heq);
    injection Heq ; intros ; subst; repeat split.
Qed.  



Lemma main_lt_rhs_aux2 : SCM.weaker
    (false, SCM.Rlf_map1 (RL1.rlf main_gt_1 rhs_main_1))
    (false, SCM.Rlf_map2 (RL2.rlf main_gt_2 rhs_main_2)).
Proof.
  unfold SCM.weaker,   main_gt_1, rhs_main_1, main_gt_2,
  rhs_main_2,
  SCM.bRlf_map1, SCM.Rlf_map1, SCM.symState_map1,
  Comp1.symState_map, Comp1.state_map,
  SCM.bRlf_map2, SCM.Rlf_map2, SCM.symState_map2,
  Comp2.symState_map, Comp2.state_map ; cbn; repeat split ;
    unfold MainMod.imp.
  * intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
              (Hc' &  Hgcd & Hx'0 & Hy'0 & Hx'y') & Heq);
      injection Heq ; intros ; subst; repeat split.
    now exists (SndCompMod.c1, x'0, y'0, x', y').
  * intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             (Hc' &  (Hx'y' & Hx')) & Heq) ;
     injection Heq ; intros ; subst; repeat split.
  now exists (FstCompMod.c1, x'0, y'0, gcd x'0 y'0, gcd x'0 y'0).
Qed.
    
Lemma main_lt_rhs_aux3 : SCM.weaker
    (false, SCM.Rlf_map2 (RL2.rlf main_lt_2 rhs_main_2))
    (false, SCM.Rlf_map1 (RL1.rlf main_lt_1 rhs_main_1)).
Proof.
  unfold SCM.weaker,   main_lt_1, rhs_main_1, main_lt_2,
  rhs_main_2,
  SCM.bRlf_map1, SCM.Rlf_map1, SCM.symState_map1,
  Comp1.symState_map, Comp1.state_map,
  SCM.bRlf_map2, SCM.Rlf_map2, SCM.symState_map2,
  Comp2.symState_map, Comp2.state_map ; cbn; repeat split ;
    unfold MainMod.imp.
  * intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             (Hc' &  Hgcd & Hx'0 & Hy'0 & Hx'y') & Heq);
      injection Heq ; intros ; subst; repeat split.
  now exists (FstCompMod.c1, x'0, y'0, x', y').
   * intros  ((((c,x0),y0), x),y)
             (((((c',x'0),y'0), x'),y') &
              (Hc' &  (Hx'y' & Hx')) & Heq) ;
      injection Heq ; intros ; subst; repeat split.
  now exists (SndCompMod.c1, x'0, y'0, gcd x'0 y'0, gcd x'0 y'0).
Qed.

Lemma main_lt_rhs_1_aux1 :
  FstCompMod.imp (FstCompMod.cnj main_lt_1 FstCompMod.final)
                 FstCompMod.bot.
Proof.
unfold  FstCompMod.imp,FstCompMod.cnj,main_lt_1,FstCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- FstCompMod.final_is_final in Hf.
unfold FstCompMod.phi in Hf ; destruct Hf ; omega.
Qed.

Definition main_lt'_1 :  SymState1 :=
  fun (s : State1) =>
    match s with (c,x0,y0,x,y) =>
         c = FstCompMod.c' /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ x < y
    end.



Lemma main_lt_rhs_1_aux2 :FstCompMod.imp (FstCompMod.symTrans main_lt_1) main_lt'_1.
Proof.
unfold FstCompMod.imp, FstCompMod.symTrans, main_lt_1, main_lt'_1.
intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; repeat split ; auto.
Qed.

Lemma main_lt_rhs_1_aux3:
FstCompMod.imp (FstCompMod.cnj main_lt'_1 FstCompMod.final)
               FstCompMod.bot.
Proof.  
unfold  FstCompMod.imp,FstCompMod.cnj,main_lt'_1,FstCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- FstCompMod.final_is_final in Hf.
unfold FstCompMod.phi in Hf ; destruct Hf; discriminate.
Qed.


Definition main_ge'_1 :  SymState1 :=
  fun (s : State1) =>
    match s with (c,x0,y0,x,y) =>
         c = FstCompMod.c' /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ y <= x
    end.
Proof.

  
Lemma main_lt_rhs_1_aux4 :
    FstCompMod.imp (FstCompMod.symTrans main_lt'_1)
    (FstCompMod.dsj (FstCompMod.dsj main_lt'_1 main_ge'_1)
                    (FstCompMod.dsj rhs_main_1 main_gt_1)).
Proof.
unfold FstCompMod.imp, FstCompMod.dsj,
  FstCompMod.symTrans, main_lt'_1, main_ge'_1,
  rhs_main_1, main_gt_1.
intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; try omega.
left. 
destruct (le_gt_dec (y' -x) x) as
      [Hdisc|Hdisc].
*right; repeat split ; auto.
  rewrite <- Hgcd.
  apply gcd_lt ; assumption.
* left; repeat split ; auto.
  rewrite <- Hgcd.
   apply gcd_lt ; assumption.
Qed.


Lemma main_lt_rhs_1_aux5:
  FstCompMod.imp (FstCompMod.cnj main_ge'_1 FstCompMod.final)
    FstCompMod.bot.
Proof.  
unfold  FstCompMod.imp,FstCompMod.cnj,main_ge'_1,FstCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- FstCompMod.final_is_final in Hf.
unfold FstCompMod.phi in Hf ; destruct Hf ;discriminate.
Qed.
  
Lemma main_lt_rhs_1_aux6:
  FstCompMod.imp (FstCompMod.symTrans main_ge'_1)
    (FstCompMod.dsj main_gt_1 rhs_main_1).
Proof.
unfold FstCompMod.imp, FstCompMod.symTrans, FstCompMod.dsj,main_ge'_1,  main_gt_1,  rhs_main_1.
intros ((((c,x0),y0), x),y)  (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; try omega.
destruct (le_gt_dec x y).
* right ; repeat split ; auto; try omega.
  assert (Heq : x  = y) ; [omega|].
  rewrite <- Hgcd; subst.
  apply gcd_eq; reflexivity.
* left; repeat split; auto.
Qed.  

         
Lemma main_lt_rhs_1 :
  RL1.proof [(false, RL1.rlf main_gt_1 rhs_main_1)]
    (false, RL1.rlf main_lt_1 rhs_main_1).
Proof.
apply RL1.Stp ; [apply main_lt_rhs_1_aux1 |].
apply RL1.Str with main_lt'_1; [apply main_lt_rhs_1_aux2 |].
apply RL1.Cof.
apply RL1.Stp; [apply main_lt_rhs_1_aux3 |].
apply RL1.Str with
          (FstCompMod.dsj
            (FstCompMod.dsj main_lt'_1 main_ge'_1)
            (FstCompMod.dsj rhs_main_1 main_gt_1))
             ; [apply main_lt_rhs_1_aux4 |].
repeat apply RL1.Spl.
* apply (RL1.Hyp []).
* apply RL1.Stp ; [apply main_lt_rhs_1_aux5 |].
  apply RL1.Str with
      (FstCompMod.dsj  main_gt_1 rhs_main_1);
    [apply  main_lt_rhs_1_aux6 |].
  apply RL1.Spl.
  + apply (RL1.Hyp [(false, RL1.rlf main_lt'_1 rhs_main_1)] []).
  + apply RL1.Trv.
*apply RL1.Trv.
*apply (RL1.Hyp [(false, RL1.rlf main_lt'_1 rhs_main_1)] []).
Qed.


Lemma main_gt_rhs_2_aux1 :
  SndCompMod.imp (SndCompMod.cnj main_gt_2 SndCompMod.final)
                 SndCompMod.bot.  

Proof.
unfold  SndCompMod.imp,SndCompMod.cnj,main_gt_2,FstCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- SndCompMod.final_is_final in Hf.
unfold SndCompMod.phi in Hf ; destruct Hf ; omega.
Qed.


Definition main_gt'_2 :  SymState2 :=
  fun (s : State2) =>
    match s with (c,x0,y0,x,y) =>
         c = SndCompMod.c'' /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ y < x
    end.



Lemma main_gt_rhs_2_aux2 :SndCompMod.imp (SndCompMod.symTrans main_gt_2) main_gt'_2.

Proof.
unfold SndCompMod.imp, SndCompMod.symTrans, main_gt_2, main_gt'_2.
intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; repeat split ; auto.
Qed.


Lemma main_gt_rhs_2_aux3:
SndCompMod.imp (SndCompMod.cnj main_gt'_2 SndCompMod.final)
               SndCompMod.bot.
Proof.  
unfold  SndCompMod.imp,SndCompMod.cnj,main_gt'_2,FstCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- SndCompMod.final_is_final in Hf.
unfold SndCompMod.phi in Hf ; destruct Hf; discriminate.
Qed.


Definition main_le'_2 :  SymState2 :=
  fun (s : State2) =>
    match s with (c,x0,y0,x,y) =>
         c = SndCompMod.c'' /\ gcd x y = gcd x0 y0 /\ 0 < x0 /\ 0 < y0 /\ x <= y
    end.
Proof.


Lemma main_gt_rhs_2_aux4 :
    SndCompMod.imp (SndCompMod.symTrans main_gt'_2)
    (SndCompMod.dsj (SndCompMod.dsj main_gt'_2 main_le'_2)
                    (SndCompMod.dsj rhs_main_2 main_lt_2)).
Proof.
unfold SndCompMod.imp, SndCompMod.dsj,
  SndCompMod.symTrans, main_gt'_2, main_le'_2,
  rhs_main_2, main_lt_2.
intros  ((((c,x0),y0), x),y)
            (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; try omega.
left. 
destruct (le_gt_dec (x' -y) y) as
      [Hdisc|Hdisc].
*right; repeat split ; auto.
  rewrite <- Hgcd.
  rewrite gcd_sym.
  rewrite gcd_lt; auto.
  apply gcd_sym.
* left; repeat split ; auto.
  rewrite <- Hgcd.
  rewrite gcd_sym.
  rewrite gcd_lt; auto.
  apply gcd_sym.
Qed.



Lemma main_gt_rhs_2_aux5:
  SndCompMod.imp (SndCompMod.cnj main_le'_2 SndCompMod.final)
    SndCompMod.bot.
Proof.  
unfold  SndCompMod.imp,SndCompMod.cnj,main_le'_2,SndCompMod.bot.
intros ((((c,x0),y0), x),y) ((Hc & Hgcd & Hx0 & Hy0 & Hxy) & Hf) ;
  subst.
rewrite <- SndCompMod.final_is_final in Hf.
unfold SndCompMod.phi in Hf ; destruct Hf ;discriminate.
Qed.



Lemma main_gt_rhs_2_aux6:
  SndCompMod.imp (SndCompMod.symTrans main_le'_2)
                 (SndCompMod.dsj main_lt_2 rhs_main_2).
Proof.
unfold SndCompMod.imp, SndCompMod.symTrans, SndCompMod.dsj,main_le'_2,  main_lt_2,  rhs_main_2.
intros ((((c,x0),y0), x),y)  (((((c',x'0),y'0), x'),y') &
             ((Hc' & Hgcd & Hx'0 & Hy'0 & Hx'y') & Htr)); subst.
inversion Htr ; subst ; clear Htr; try omega.
destruct (le_gt_dec y x).

* right ; repeat split ; auto; try omega.
  assert (Heq : x  = y) ; [omega|].
  rewrite <- Hgcd; subst.
  apply gcd_eq; reflexivity.
* left; repeat split; auto.
Qed.  
  
Lemma  main_gt_rhs_2 :
  RL2.proof [(false, RL2.rlf main_lt_2 rhs_main_2)]
    (false, RL2.rlf main_gt_2 rhs_main_2).
Proof.
apply RL2.Stp ; [apply main_gt_rhs_2_aux1 |].
apply RL2.Str with main_gt'_2; [apply main_gt_rhs_2_aux2 |].
apply RL2.Cof.
apply RL2.Stp; [apply main_gt_rhs_2_aux3 |].
apply RL2.Str with
          (SndCompMod.dsj
            (SndCompMod.dsj main_gt'_2 main_le'_2)
            (SndCompMod.dsj rhs_main_2 main_lt_2))
             ; [apply main_gt_rhs_2_aux4 |].
repeat apply RL2.Spl.
* apply (RL2.Hyp []).
* apply RL2.Stp ; [apply main_gt_rhs_2_aux5 |].
  apply RL2.Str with
      (SndCompMod.dsj  main_lt_2 rhs_main_2);
    [apply  main_gt_rhs_2_aux6 |].
  apply RL2.Spl.
  + apply (RL2.Hyp [(false, RL2.rlf main_gt'_2 rhs_main_2)] []).
  + apply RL2.Trv.
*apply RL2.Trv.
*apply (RL2.Hyp [(false, RL2.rlf main_gt'_2 rhs_main_2)] []).
Qed.

Lemma main_lt_rhs : RL.proof []   (true, RL.rlf main_lt rhs_main).
Proof.
apply RL.proof_mono_bool.
eapply SCM.ProofComp_gen with
    (H := []) (H1 := []) (H2 := [])
    (f1 :=  RL1.rlf main_lt_1 rhs_main_1)
    (f'2 :=  RL1.rlf main_gt_1 rhs_main_1)
    (f2 :=  RL2.rlf main_gt_2 rhs_main_2)
    (f'1 :=  RL2.rlf main_lt_2 rhs_main_2) 
;  try apply SCM.weakerH_refl.
* left.
  apply main_lt_rhs_aux1.
* apply main_lt_rhs_aux2.
* apply main_lt_rhs_aux3.
* apply main_lt_rhs_1.
* apply main_gt_rhs_2.
Qed.


Lemma main_gt_rhs : RL.proof []   (true, RL.rlf main_gt rhs_main).
apply RL.proof_mono_bool.
eapply SCM.ProofComp_gen with
    (H := []) (H1 := []) (H2 := [])
    (f1 :=  RL1.rlf main_lt_1 rhs_main_1)
    (f2 :=  RL2.rlf main_gt_2 rhs_main_2)
    (f'1 :=  RL2.rlf main_lt_2 rhs_main_2)
    (f'2 :=  RL1.rlf main_gt_1 rhs_main_1)   
;  try apply SCM.weakerH_refl.
* right.
  apply main_gt_rhs_aux1.
* apply main_lt_rhs_aux2.
* apply main_lt_rhs_aux3.
* apply main_lt_rhs_1.
* apply main_gt_rhs_2.
Qed.

Lemma main_aux2 : MainMod.imp (MainMod.symTrans lhs_main)
    (MainMod.dsj
       (MainMod.dsj main_lt rhs_main)
       main_gt).
unfold MainMod.imp, MainMod.symTrans, MainMod.dsj, lhs_main, main_lt, rhs_main, main_gt.
intros ((((c,x0),y0), x),y) (((((c',x'0),y'0), x'),y') & ((Hc' & Hx'0 & Hy'0) & Htr)) ;
  subst.
inversion Htr ; subst ; clear Htr.
destruct (lt_eq_lt_dec x y) as [[Hxy|Hxy]|Hxy].
 *  left; left; repeat split; auto.
 *  left; right; repeat split; subst ;auto.
    apply gcd_eq; reflexivity.
 *   right ; repeat split; auto.
Qed.



Lemma  main : RL.valid (RL.rlf lhs_main rhs_main).
Proof.
  apply RL.soundness.

  apply RL.Stp.
  * apply main_aux1.
  * apply RL.Str with (MainMod.dsj (MainMod.dsj main_lt rhs_main) main_gt).
    + apply main_aux2.
    + repeat apply RL.Spl.
     - apply main_lt_rhs.
     - apply RL.Trv.
     - apply main_gt_rhs.
Qed.

(* main result spelled out without intermediary definitions *)

Theorem Main : RL.valid  (RL.rlf
    (fun s =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c0 /\ 0 < x0 /\ 0 < y0
    end)
  (fun s  =>
    match s with (c,x0,y0,x,y) =>
         c = MainMod.c1 /\ x = y /\ x = gcd x0 y0
    end)).
Proof.
  apply main.
Qed.

End ProofMod.
