Require Import SigTS.
Require Import SigComp.
Require Import ReachabilityLogic.
Require Import Coq.Program.Equality.
Require Import Classical.
Require Import List.
Import ListNotations.



Module LogicStructCompMod
       (TS1 TS2 TS:SigTSMod)
       (Comp1: SigCompMod TS TS1)
       (Comp2: SigCompMod TS TS2).
      
Module RL1 := ReachabilityLogicMod TS1.
Module RL2 := ReachabilityLogicMod TS2.
Module RL :=  ReachabilityLogicMod TS.


Definition symState_map1 := Comp1.symState_map.
Definition symState_map2 := Comp2.symState_map.


  
Definition Rlf_map1 : RL1.Rlf -> RL.Rlf
  := fun f => 
       (RL.rlf (symState_map1 (RL1.lhs f) )
                     (symState_map1 (RL1.rhs f))).



Definition Rlf_map2 : RL2.Rlf -> RL.Rlf
  := fun f => 
       (RL.rlf (symState_map2 (RL2.lhs f) )
                     (symState_map2 (RL2.rhs f))).


Definition bRlf_map1 :(bool * RL1.Rlf)->(bool * RL.Rlf)
  := fun bf =>  match bf with (b,f') =>
                              (b, Rlf_map1 f') end.


Definition bRlf_map2 :(bool * RL2.Rlf)->(bool * RL.Rlf)
  := fun bf =>  match bf with (b,f') =>
                              (b, Rlf_map2 f') end.


Lemma symState_map_dsj1 : forall l1 l2,
    TS.imp (symState_map1 (TS1.dsj l1 l2))
           (TS.dsj (symState_map1 l1) (symState_map1 l2)).
Proof.
  intros.
  unfold TS.imp, TS1.dsj, TS.dsj, symState_map1,Comp1.symState_map.
  intros s (s' & Hdsj & Heq) ; subst.
  destruct Hdsj.
  * left ;  now exists s'.
  * right ; now exists s'.
Qed.

Lemma symState_map_dsj2 : forall l1 l2,
    TS.imp (symState_map2 (TS2.dsj l1 l2))
           (TS.dsj (symState_map2 l1) (symState_map2 l2)).
Proof.
  intros.
  unfold TS.imp, TS2.dsj, TS.dsj, symState_map2,Comp2.symState_map.
  intros s (s' & Hdsj & Heq) ; subst.
  destruct Hdsj.
  * left ;  now exists s'.
  * right ; now exists s'.
Qed.


Lemma final1_final :  forall l, TS1.imp (TS1.cnj l TS1.final) TS1.bot ->
  TS.imp (TS.cnj (symState_map1 l) TS.final) TS.bot.
Proof.
intro l.
unfold TS1.imp, TS1.cnj, TS1.final, TS1.bot, TS.imp, TS.cnj,
symState_map1, Comp1.symState_map, TS.final, TS.bot.  
intros H'imp s  ((s' & Hls' & Heq) & Hfin) ; subst.
apply (H'imp s') ; repeat split ; auto.
intros s Htr.
apply (Hfin (Comp1.state_map s)).
apply Comp1.subtrans ; assumption.
Qed.


Lemma final2_final :  forall l, TS2.imp (TS2.cnj l TS2.final) TS2.bot ->
  TS.imp (TS.cnj (symState_map2 l) TS.final) TS.bot.
Proof.
intro l.
unfold TS2.imp, TS2.cnj, TS2.final, TS2.bot, TS.imp, TS.cnj,
symState_map2, Comp2.symState_map, TS.final, TS.bot.  
intros H'imp s  ((s' & Hls' & Heq) & Hfin) ; subst.
apply (H'imp s') ; repeat split ; auto.
intros s Htr.
apply (Hfin (Comp2.state_map s)).
apply Comp2.subtrans ; assumption.
Qed.
  
Lemma symTrans_map1  : forall l, TS1.imp (TS1.cnj l TS1.final)
        TS1.bot -> TS.imp (TS.symTrans (symState_map1 l))
                    (symState_map1 (TS1.symTrans l)).
intros l.
unfold TS1.imp, TS1.cnj, TS1.bot, TS.imp, TS.symTrans,
symState_map1, Comp1.symState_map, TS1.symTrans.
intros Hcnj s (s1 & (s' & Hls' & Heq ) & Htr) ; subst.
destruct (classic (Comp1.isImage s)).
* destruct H as (s2' & Heq) ; subst.
  exists s2';  split ; auto.
  exists s' ;  split ; auto.
  eapply Comp1.full_subtrans; eauto.
* 
  exfalso.
  apply (Hcnj s') ; split; auto.
  eapply Comp1.clean_border ; eauto.
Qed.


Lemma symTrans_map2  : forall l, TS2.imp (TS2.cnj l TS2.final)
        TS2.bot -> TS.imp (TS.symTrans (symState_map2 l))
                    (symState_map2 (TS2.symTrans l)).
intros l.
unfold TS2.imp, TS2.cnj, TS2.bot, TS.imp, TS.symTrans,
symState_map2, Comp2.symState_map, TS2.symTrans.
intros Hcnj s (s2 & (s' & Hls' & Heq ) & Htr) ; subst.
destruct (classic (Comp2.isImage s)).
* destruct H as (s2' & Heq) ; subst.
  exists s2';  split ; auto.
  exists s' ;  split ; auto.
  eapply Comp2.full_subtrans; eauto.
* 
  exfalso.
  apply (Hcnj s') ; split; auto.
  eapply Comp2.clean_border ; eauto.
Qed.



 Lemma ProofComp_aux1 :
   forall H b f',  RL1.proof H (b,f') ->
                   RL.proof (map bRlf_map1 H) (b,Rlf_map1 f').
Proof.   
intros H b f' Hproof.
 (* remember (b, f') as bf' eqn: Hbf'.$ *)
 (* pattern b, f', (map bRlf_map H), (b, Rlf_map f').*)
 (* remember (map bRlf_map H) as mapbRlf eqn: HmapbRlf.
  remember (b, Rlf_map f') as bRlf eqn: HbRlf. *)
dependent induction Hproof.
* apply RL.Trv.
* repeat rewrite map_app.
  eapply RL.Hyp.
* apply RL.Str  with (symState_map1 l') .
  + unfold TS.imp, symState_map1, Comp1.symState_map.
    intros s (s' & Hl & Heq); subst.
    exists s' ; repeat split ; auto.
  + apply (IHHproof _ _ (eq_refl _)).
* apply RL.Str with (TS.dsj(symState_map1 l1) (symState_map1 l2)).
  + apply symState_map_dsj1.  
  +  apply RL.Spl.
    - apply (IHHproof1 _ _ (eq_refl _)).
    - apply (IHHproof2 _ _ (eq_refl _)).
* apply RL.Tra with (symState_map1 m).
  + apply  (IHHproof1 _ _ (eq_refl _)).
  + apply  (IHHproof2 _ _ (eq_refl _)).
* apply RL.Stp.
  +  unfold TS.imp.
     apply (final1_final _ i).
  + eapply RL.Str  with (symState_map1 (TS1.symTrans l)).    
   - apply (symTrans_map1 _ i). 
   - apply (IHHproof _ _ (eq_refl _)).
* apply RL.Cof.
  apply  (IHHproof _ _ (eq_refl _)).
* eapply RL.Cut.
   + apply  (IHHproof1 _ _ (eq_refl _)).
   + apply  (IHHproof2 _ _ (eq_refl _)).
* repeat rewrite map_app. 
  apply RL.Clr ; cbn.
  repeat rewrite <- map_app.
  apply  (IHHproof _ _ (eq_refl _)).
Qed.



Lemma ProofComp_aux2 :
   forall H b f',  RL2.proof H (b,f') ->
                   RL.proof (map bRlf_map2 H) (b,Rlf_map2 f').
Proof.   
intros H b f' Hproof.
dependent induction Hproof.
*  apply RL.Trv.
* repeat rewrite map_app.
  eapply RL.Hyp.
*  apply RL.Str  with (symState_map2 l') .
  + unfold TS.imp, symState_map2, Comp2.symState_map.
    intros s (s' & Hl & Heq); subst.
    exists s' ; repeat split ; auto.
  + apply (IHHproof _ _ (eq_refl _)).
* apply RL.Str with (TS.dsj(symState_map2 l1) (symState_map2 l2)).
  + apply symState_map_dsj2.  
  + apply RL.Spl.
    - apply (IHHproof1 _ _ (eq_refl _)).
    - apply (IHHproof2 _ _ (eq_refl _)).
* apply RL.Tra with (symState_map2 m).
  + apply  (IHHproof1 _ _ (eq_refl _)).
  + apply  (IHHproof2 _ _ (eq_refl _)).
* apply RL.Stp.
  +  unfold TS.imp.
     apply (final2_final _ i).
  + eapply RL.Str  with (symState_map2 (TS2.symTrans l)).
   - apply (symTrans_map2 _ i). 
   -  apply (IHHproof _ _ (eq_refl _)).
* apply RL.Cof.
  apply  (IHHproof _ _ (eq_refl _)).
* eapply RL.Cut.
   + apply  (IHHproof1 _ _ (eq_refl _)).
   + apply  (IHHproof2 _ _ (eq_refl _)).
* repeat rewrite map_app. 
  apply RL.Clr ; cbn.
  repeat rewrite <- map_app.
  apply  (IHHproof _ _ (eq_refl _)).
Qed.

Definition weaker := RL.weaker.
Definition weakerH := RL.weakerH.

Lemma weaker_refl : forall bf,  weaker bf bf.
Proof.
  intros (b, (l,r)) ; unfold weaker ;
    repeat split ; apply RL.imp_refl.
Qed.

Lemma weakerH_refl : forall H,  weakerH H H.
Proof.
intros H n bf Hnth.
exists n, bf ; split ; auto.
apply weaker_refl.
Qed.

Lemma weaker_weakerH : forall bf1 bf2 H1 H2,
    weaker bf1 bf2 ->weakerH H1 H2 ->
    weakerH (bf1::H1) (bf2::H2).
Proof.
  intros bf1 bf2 H1 H2 Hw HwH n bf Hnth.
  destruct n.
  * injection Hnth ; intros ; subst.
    exists 0, bf2; split; auto.
  * destruct (HwH _ _ Hnth) as (m & bf' & Hnth' & Hw').
    exists (S m), bf'; split ; auto.
Qed.

Theorem ProofComp_gen :
  forall (f1 f'2: RL1.Rlf)(H1 : list (bool * RL1.Rlf))
         (f2 f'1 :RL2.Rlf)(H2 : list (bool * RL2.Rlf))
         (f :RL.Rlf) (H : list (bool * RL.Rlf)),
    weakerH (map bRlf_map1 H1) H ->
    weakerH (map bRlf_map2 H2) H ->
    weakerH (map bRlf_map1 H1) (map bRlf_map2 H2) ->
    weakerH (map bRlf_map2 H2) (map bRlf_map1 H1) ->
    (weaker (false,f) (bRlf_map1 (false,f1)) +
    weaker (false,f) (bRlf_map2 (false,f2))) ->
    weaker  (false,(Rlf_map1 f'2))(false, (Rlf_map2 f2)) ->
    weaker  (false,(Rlf_map2 f'1))(false, (Rlf_map1 f1)) ->
    RL1.proof ((false,f'2) :: H1)
                 (false,f1) ->
    RL2.proof ((false,f'1) :: H2)
                            (false,f2) ->
    RL.proof H (false,f).
Proof.
intros f1 f'2 H1 f2 f'1 H2 f h HwH1 HwH2 Hwmap1 Hwmap2
         [Hplus| Hplus] Hw2 Hw1 Hpr1 Hpr2.
* eapply RL.proof_mono_goals; eauto.
  eapply RL.proof_mono_hypos; eauto.
  apply RL.MutCompStrat with (Rlf_map2 f2).
  + apply RL.proof_mono_hypos with
      (map bRlf_map2 ((false, f'1) :: H2)).
    - apply ProofComp_aux2 ; assumption.
    - apply weaker_weakerH; auto.
      apply weakerH_refl.
 + apply RL.proof_mono_hypos with
     (map bRlf_map1 ((false, f'2) :: H1)).
    - apply ProofComp_aux1 ; assumption.
    - apply weaker_weakerH ; assumption.
* eapply RL.proof_mono_goals; eauto.
  eapply RL.proof_mono_hypos; eauto.
  apply RL.MutCompStrat with (Rlf_map1 f1).
  + apply RL.proof_mono_hypos with
      (map bRlf_map1 ((false, f'2) :: H1)).
    - apply ProofComp_aux1 ; assumption.
    - apply weaker_weakerH; assumption.
 + eapply RL.proof_mono_hypos with
     (map bRlf_map2 ((false, f'1) :: H2)).
    - apply ProofComp_aux2 ; assumption.
    - apply weaker_weakerH ; auto.
      apply weakerH_refl.    
Qed.


End LogicStructCompMod.
