Require Import SigTS.
Require Import SigComp.
Require Export ReachabilityLogic.
Require Import Classical.


Module StructCompMod (TS:SigTSMod)(TS':SigTSMod)
                     (Comp: SigCompMod TS TS').
  
Module RL:= ReachabilityLogicMod TS.
Module RL' := ReachabilityLogicMod TS'.
Import Comp.

(*Definition is_image(s:TS.State) : Prop := exists s',  s = state_map s'.*)


  
Definition symState_map : RL'.SymState -> RL.SymState  :=
     fun (p:RL'.SymState)(s:TS.State) => exists s',  p s' /\ (s = state_map s').

Lemma path_rev_map :
  forall (tau:RL.seq),
   RL.isPath tau -> RL.isFinite tau ->
    forall (l':RL'.SymState), (symState_map l') (RL.head tau) ->
     exists tau',RL'.isPath tau' /\ RL'.isFinite tau' /\ l'(RL'.head tau') /\
      forall (n' :nat) (sn':TS'.State), (RL'.nth n' tau') = Some sn' ->
         exists sn, (RL.nth n' tau) = Some sn /\ sn = state_map sn'.
Proof.
   intros tau HisPath HisFinite.
   induction HisFinite ; cbn in * ; intros l' Hini.
   * destruct Hini as ( s' & Hls' & Heq) ; subst.
     exists (RL'.one s') ; repeat split ; auto.
     + constructor.
       inversion HisPath ; subst.
       apply TS'.final_is_final.
       intros s''  Habs'.
       rewrite TS.final_is_final in H0.
       apply (H0 (state_map s'')).
       apply subtrans ; auto.
     + constructor.
     + intros n' sn' Hnth.
       exists (state_map sn') ; split ; auto.
       destruct n' ; cbn in *.
       - injection Hnth ; intros; subst ; auto.
       -  discriminate.
   * assert (HisPath' : RL.isPath sq); [inversion HisPath ; auto | ].
     specialize (IHHisFinite HisPath').
     assert (H: isImage state_map s).
     {destruct Hini as (s'' & Hl' & Heq') ; subst.
      now exists s''. }
     destruct H as (s' & Hs').
     inversion HisPath  ; subst.
     destruct (classic (isImage state_map (RL.head sq))).
     +  destruct H as (hdsq' & Hhdsq').
        rewrite  Hhdsq' in H1.
        generalize (full_subtrans _ _ _ _  H1 (eq_refl _) (eq_refl _)) ;
          intro Hfull.
                       
       assert(HsymTrans : (symState_map (RL'.symTrans l')) (RL.head sq)).
     {
       exists hdsq' ; split ; subst ; auto.
       exists s'; repeat split ; auto.
       inversion Hini as (x & Hc1 & Hc2) ; subst.
       apply state_map_inj in Hc2 ; subst ; auto.       
     }  
     specialize (IHHisFinite _ HsymTrans).
     destruct IHHisFinite as (tau' & HisPath'' & HisFinite'' & Hini'' & Hforall).
     exists (RL'.add s' tau') ; repeat split.
       - constructor ; auto.
         destruct tau', sq.
         ** 
            specialize (Hforall 0 s (eq_refl _) ) ; cbn in *.
            destruct Hforall as (s0' & Heq' & Heq'')  ; subst.
            injection Heq'; intros ; subst.
            apply state_map_inj in H ; subst ; auto.
         **  
             specialize (Hforall 0 s (eq_refl _) ) ; cbn in *.
             destruct Hforall as (s0' & Heq' & Heq'')  ; subst.
             injection Heq'; intros ; subst.
             apply state_map_inj in H ; subst ; auto.
         ** 
            specialize (Hforall 0 s (eq_refl _) ) ; cbn in *.
            destruct Hforall as (s0' & Heq' & Heq'')  ; subst.
            injection Heq'; intros ; subst.
            apply state_map_inj in H ; subst ; auto.
         ** specialize (Hforall 0 s (eq_refl _) ) ; cbn in *.
            destruct Hforall as (s0' & Heq' & Heq'')  ; subst.
            injection Heq'; intros ; subst.
            apply state_map_inj in H ; subst ; auto.
       - constructor ; auto.
       - cbn.
         destruct Hini as (s'' & Hl' & Heq').
         apply state_map_inj in Heq' ; subst ; auto.
       - intros n' sn' Hnth.
         destruct n' ; cbn in *.
         ** injection Hnth ; intros ; subst.
            now exists (state_map sn').
         ** apply Hforall ; auto.
   + unfold isImage in H.
     generalize (clean_border _ _ H1 H); intro Hcb.
     assert (Hfinl : TS'.final s').
     {
       apply TS'.final_is_final.
        intros s'0 Htrans.
        rewrite TS'.final_is_final in Hcb.
        apply (Hcb s'0); auto.
     }
     clear H Hcb.
     exists (RL'.one s') ; repeat split.
     - constructor; auto.
     - constructor.
     -  cbn.
        destruct Hini as (s'' & Hl' & Heq') ; subst.
        apply state_map_inj in Heq' ; subst ; auto.
     - intros.
       destruct n'; try discriminate.
       cbn in *.
       injection H ; intros ; subst.
       now exists (state_map sn').
Qed.

Definition Rlf_map : RL'.Rlf -> RL.Rlf
  := fun f =>  match f with (RL'.rlf l' r') =>
  (RL.rlf (symState_map l') (symState_map r')) end.

Theorem ValidComp : forall f', RL'.valid f' -> RL.valid (Rlf_map f').
  Proof.  
  intros f' Hval'.  
  destruct f' as (l',r')  ; cbn.
  rewrite RL'.valid_iff_finite_ex in Hval'.
  rewrite RL.valid_iff_finite_ex.
  intros tau Hinit HisPath HisFinite.
  destruct  (path_rev_map _ HisPath HisFinite _ Hinit) as (tau' & HisPath' & HisFinite' & Hinit' & Hrest').
  destruct ( Hval' _ Hinit' HisPath' HisFinite') as (n & s & Hsome & Hr') .
  destruct (Hrest' _ _ Hsome) as (s0 &  Hnth& Hs0).
  exists n, s0 ; split ; auto ; subst.
  unfold symState_map.
  now exists s.
 Qed.                            

End StructCompMod.
